#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['oxlang']

package_data = \
{'': ['*'],
 'oxlang': ['.mypy_cache/*',
            '.mypy_cache\\3.9/*',
            '.mypy_cache\\3.9\\_typeshed/*',
            '.mypy_cache\\3.9\\collections/*',
            '.mypy_cache\\3.9\\importlib/*',
            '.mypy_cache\\3.9\\os/*',
            '.mypy_cache\\3.9\\oxlang/*',
            'stdlib/*']}

install_requires = \
['sly>=0.4']

extras_require = \
{'test': ['pytest>=6.2.2']}

entry_points = \
{'console_scripts': ['ox = oxlang.repl:main']}

setup(name='oxlang',
      version='0.0.1a0',
      description='Python implementation of ox.',
      author='Ong Yong Xin',
      author_email='ongyongxin2020+github@gmail.com',
      url='https://github.com/ongyx/ox',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.6',
     )
