# coding: utf8
"""Python implementation of ox.
Inspired by the Cub Programming Language created by Louis D'hauwe.
"""

from .__version__ import __version__  # noqa: F401
from .runtime import Runtime  # noqa: F401
